@@ -1,29 +1,31 @@
 #!/usr/bin/env python
-# vim:fileencoding=utf-8
 # License: GPLv3 Copyright: 2013, Kovid Goyal <kovid at kovidgoyal.net>
 
 
 import weakref
 from qt.core import (
-    QApplication, QByteArray, QCalendarWidget, QCheckBox, QColor, QColorDialog, QFrame,
+    QApplication, QBrush, QByteArray, QCalendarWidget, QCheckBox, QColor, QColorDialog,
     QComboBox, QDate, QDateTime, QDateTimeEdit, QDialog, QDialogButtonBox, QFont,
-    QFontInfo, QFontMetrics, QIcon, QKeySequence, QLabel, QLayout, QMenu, QMimeData,
-    QPalette, QPixmap, QPoint, QPushButton, QRect, QScrollArea, QSize, QSizePolicy,
-    QStyle, QStyledItemDelegate, Qt, QTabWidget, QTextBrowser, QToolButton, QTextCursor,
-    QUndoCommand, QUndoStack, QUrl, QWidget, pyqtSignal
+    QFontInfo, QFontMetrics, QFrame, QIcon, QKeySequence, QLabel, QLayout, QMenu,
+    QMimeData, QPainter, QPalette, QPixmap, QPoint, QPushButton, QRect, QScrollArea,
+    QSize, QSizePolicy, QStyle, QStyledItemDelegate, QStyleOptionToolButton,
+    QStylePainter, Qt, QTabWidget, QTextBrowser, QTextCursor, QTimer, QToolButton,
+    QUndoCommand, QUndoStack, QUrl, QWidget, pyqtSignal,
 )
 
+from calibre import prepare_string_for_xml
+from calibre.constants import builtin_colors_dark, builtin_colors_light
 from calibre.ebooks.metadata import rating_to_stars
 from calibre.gui2 import UNDEFINED_QDATETIME, gprefs, rating_font
 from calibre.gui2.complete2 import EditWithComplete, LineEdit
 from calibre.gui2.widgets import history
 from calibre.utils.config_base import tweaks
 from calibre.utils.date import UNDEFINED_DATE
-from polyglot.builtins import unicode_type
+from calibre.utils.localization import _
 from polyglot.functools import lru_cache
 
 
-class HistoryMixin(object):
+class HistoryMixin:
 
     max_history_items = None
     min_history_entry_length = 3
@@ -50,7 +52,7 @@ class HistoryMixin(object):
         return history.get(self.store_name, [])
 
     def save_history(self):
-        ct = unicode_type(self.text())
+        ct = str(self.text())
         if len(ct) >= self.min_history_entry_length:
             try:
                 self.history.remove(ct)
@@ -104,7 +106,7 @@ class ColorButton(QPushButton):
 
     @color.setter
     def color(self, val):
-        val = unicode_type(val or '')
+        val = str(val or '')
         col = QColor(val)
         orig = self._color
         if col.isValid():
@@ -123,7 +125,7 @@ class ColorButton(QPushButton):
     def choose_color(self):
         col = QColorDialog.getColor(QColor(self._color or Qt.GlobalColor.white), self, _('Choose a color'))
         if col.isValid():
-            self.color = unicode_type(col.name())
+            self.color = str(col.name())
 
 
 def access_key(k):
@@ -159,6 +161,41 @@ class RightClickButton(QToolButton):
         return QToolButton.mousePressEvent(self, ev)
 
 
+class CenteredToolButton(RightClickButton):
+
+    def __init__(self, icon, text, parent=None):
+        super().__init__(parent)
+        self.setText(text)
+        self.setIcon(icon)
+        self.setSizePolicy(QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed))
+        self.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
+        self.text_flags =  Qt.TextFlag.TextSingleLine | Qt.AlignmentFlag.AlignCenter
+
+    def paintEvent(self, ev):
+        painter = QStylePainter(self)
+        opt = QStyleOptionToolButton()
+        self.initStyleOption(opt)
+        text = opt.text
+        opt.text = ''
+        opt.icon = QIcon()
+        s = painter.style()
+        painter.drawComplexControl(QStyle.ComplexControl.CC_ToolButton, opt)
+        if s.styleHint(QStyle.StyleHint.SH_UnderlineShortcut, opt, self):
+            flags = self.text_flags | Qt.TextFlag.TextShowMnemonic
+        else:
+            flags = self.text_flags | Qt.TextFlag.TextHideMnemonic
+        fw = s.pixelMetric(QStyle.PixelMetric.PM_DefaultFrameWidth, opt, self)
+        opt.rect.adjust(fw, fw, -fw, -fw)
+        w = opt.iconSize.width()
+        text_rect = opt.rect.adjusted(w, 0, 0, 0)
+        painter.drawItemText(text_rect, flags, opt.palette, self.isEnabled(), text)
+        fm = QFontMetrics(opt.font)
+        text_rect = s.itemTextRect(fm, text_rect, flags, self.isEnabled(), text)
+        left = text_rect.left() - w - 4
+        pixmap_rect = QRect(left, opt.rect.top(), opt.iconSize.width(), opt.rect.height())
+        painter.drawItemPixmap(pixmap_rect, Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter, self.icon().pixmap(opt.iconSize))
+
+
 class Dialog(QDialog):
 
     '''
@@ -172,34 +209,35 @@ class Dialog(QDialog):
     for the first time.
     '''
 
-    def __init__(self, title, name, parent=None, prefs=gprefs):
+    def __init__(
+            self, title,
+            name, parent=None, prefs=gprefs,
+            default_buttons=QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
+    ):
         QDialog.__init__(self, parent)
         self.prefs_for_persistence = prefs
         self.setWindowTitle(title)
         self.name = name
-        self.bb = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
+        self.bb = QDialogButtonBox(default_buttons)
         self.bb.accepted.connect(self.accept)
         self.bb.rejected.connect(self.reject)
 
         self.setup_ui()
 
-        self.resize(self.sizeHint())
-        geom = self.prefs_for_persistence.get(name + '-geometry', None)
-        if geom is not None:
-            QApplication.instance().safe_restore_geometry(self, geom)
+        self.restore_geometry(self.prefs_for_persistence, self.name + '-geometry')
         if hasattr(self, 'splitter'):
-            state = self.prefs_for_persistence.get(name + '-splitter-state', None)
+            state = self.prefs_for_persistence.get(self.name + '-splitter-state', None)
             if state is not None:
                 self.splitter.restoreState(state)
 
     def accept(self):
-        self.prefs_for_persistence.set(self.name + '-geometry', bytearray(self.saveGeometry()))
+        self.save_geometry(self.prefs_for_persistence, self.name + '-geometry')
         if hasattr(self, 'splitter'):
             self.prefs_for_persistence.set(self.name + '-splitter-state', bytearray(self.splitter.saveState()))
         QDialog.accept(self)
 
     def reject(self):
-        self.prefs_for_persistence.set(self.name + '-geometry', bytearray(self.saveGeometry()))
+        self.save_geometry(self.prefs_for_persistence, self.name + '-geometry')
         if hasattr(self, 'splitter'):
             self.prefs_for_persistence.set(self.name + '-splitter-state', bytearray(self.splitter.saveState()))
         QDialog.reject(self)
@@ -356,7 +394,7 @@ class FlowLayout(QLayout):  # {{{
         self.do_layout(rect, apply_geometry=True)
 
     def expandingDirections(self):
-        return Qt.Orientations(0)
+        return Qt.Orientation(0)
 
     def minimumSize(self):
         size = QSize()
@@ -440,6 +478,43 @@ class FlowLayout(QLayout):  # {{{
 # }}}
 
 
+class Separator(QWidget):  # {{{
+
+    ''' Vertical separator lines usable in FlowLayout '''
+
+    def __init__(self, parent, widget_for_height=None):
+        '''
+        You must provide a widget in the layout either here or with setBuddy.
+        The height of the separator is computed using this widget,
+        '''
+        QWidget.__init__(self, parent)
+        self.bcol = QApplication.instance().palette().color(QPalette.ColorRole.Text)
+        self.update_brush()
+        self.widget_for_height = widget_for_height
+        self.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.MinimumExpanding)
+
+    def update_brush(self):
+        self.brush = QBrush(self.bcol)
+        self.update()
+
+    def setBuddy(self, widget_for_height):
+        ''' See __init__. This is repurposed to support Qt Designer .ui files. '''
+        self.widget_for_height = widget_for_height
+
+    def sizeHint(self):
+        return QSize(1, 1 if self.widget_for_height is None else self.widget_for_height.height())
+
+    def paintEvent(self, ev):
+        painter = QPainter(self)
+        # Purely subjective: shorten the line a bit to look 'better'
+        r = ev.rect()
+        r.setTop(r.top() + 3)
+        r.setBottom(r.bottom() - 3)
+        painter.fillRect(r, self.brush)
+        painter.end()
+# }}}
+
+
 class HTMLDisplay(QTextBrowser):
 
     anchor_clicked = pyqtSignal(object)
@@ -455,7 +530,7 @@ class HTMLDisplay(QTextBrowser):
         f = QFontInfo(font)
         delta = tweaks['change_book_details_font_size_by'] + 1
         if delta:
-            font.setPixelSize(f.pixelSize() + delta)
+            font.setPixelSize(int(f.pixelSize() + delta))
             self.setFont(font)
         self.setFrameShape(QFrame.Shape.NoFrame)
         self.setOpenLinks(False)
@@ -497,9 +572,9 @@ class HTMLDisplay(QTextBrowser):
         if qurl.isLocalFile():
             path = qurl.toLocalFile()
             try:
-                with lopen(path, 'rb') as f:
+                with open(path, 'rb') as f:
                     data = f.read()
-            except EnvironmentError:
+            except OSError:
                 if path.rpartition('.')[-1].lower() in {'jpg', 'jpeg', 'gif', 'png', 'bmp', 'webp'}:
                     return QByteArray(bytearray.fromhex(
                         '89504e470d0a1a0a0000000d49484452'
@@ -509,6 +584,8 @@ class HTMLDisplay(QTextBrowser):
                         '426082'))
             else:
                 return QByteArray(data)
+        elif qurl.scheme() == 'calibre-icon':
+            return QIcon.icon_as_png(qurl.path().lstrip('/'), as_bytearray=True)
         else:
             return QTextBrowser.loadResource(self, rtype, qurl)
 
@@ -527,7 +604,7 @@ class ScrollingTabWidget(QTabWidget):
             # widgets added to a tab widget, which looks horrible.
             if (cm.left(), cm.top(), cm.right(), cm.bottom()) == (0, 0, 0, 0):
                 pl.setContentsMargins(9, 9, 9, 9)
-        name = 'STW{}'.format(abs(id(self)))
+        name = f'STW{abs(id(self))}'
         sw.setObjectName(name)
         sw.setWidget(page)
         sw.setWidgetResizable(True)
@@ -649,10 +726,67 @@ class DateTimeEdit(QDateTimeEdit):
             return QDateTimeEdit.keyPressEvent(self, ev)
 
 
+class MessagePopup(QLabel):
+
+    undo_requested = pyqtSignal(object)
+    OFFSET_FROM_TOP = 25
+
+    def __init__(self, parent):
+        QLabel.__init__(self, parent)
+        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
+        self.undo_data = None
+        if QApplication.instance().is_dark_theme:
+            c = builtin_colors_dark['green']
+        else:
+            c = builtin_colors_light['green']
+        self.color = self.palette().color(QPalette.ColorRole.WindowText).name()
+        bg = QColor(c).getRgb()
+        self.setStyleSheet(f'''QLabel {{
+            background-color: rgba({bg[0]}, {bg[1]}, {bg[2]}, 0.85);
+            border-radius: 4px;
+            color: {self.color};
+            padding: 0.5em;
+        }}'''
+        )
+        self.linkActivated.connect(self.link_activated)
+        self.close_timer = t = QTimer()
+        t.setSingleShot(True)
+        t.timeout.connect(self.hide)
+        self.setMouseTracking(True)
+        self.hide()
+
+    def mouseMoveEvent(self, ev):
+        self.close_timer.start()
+        return super().mouseMoveEvent(ev)
+
+    def link_activated(self, link):
+        self.hide()
+        if link.startswith('undo://'):
+            self.undo_requested.emit(self.undo_data)
+
+    def __call__(self, text='Testing message popup', show_undo=True, timeout=5000, has_markup=False):
+        text = '<p>' + (text if has_markup else prepare_string_for_xml(text))
+        if show_undo:
+            self.undo_data = show_undo
+            text += '\xa0\xa0<a style="text-decoration: none" href="undo://me.com">{}</a>'.format(_('Undo'))
+        text += f'\xa0\xa0<a style="text-decoration: none; color: {self.color}" href="close://me.com">✖</a>'
+        self.setText(text)
+        self.resize(self.sizeHint())
+        self.position_in_parent()
+        self.show()
+        self.raise_()
+        self.close_timer.start(timeout)
+
+    def position_in_parent(self):
+        p = self.parent()
+        self.move((p.width() - self.width()) // 2, self.OFFSET_FROM_TOP)
+
+
+
 if __name__ == '__main__':
     from calibre.gui2 import Application
     app = Application([])
     app.load_builtin_fonts()
     w = RatingEditor.test()
     w.show()
-    app.exec_()
+    app.exec()
